/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import org.sablecc.sablecc.analysis.DepthFirstAdapter;
import org.sablecc.sablecc.node.*;

public class GenerateAlternativeCodeForParser extends DepthFirstAdapter
{
  String currentAlt;
  String realcurrentAlt;
  BufferedWriter file;
  private File pkgDir;

  private ResolveTransformIds transformIds;
  private ComputeCGNomenclature CG;
  private ComputeSimpleTermPosition CTP;
  private MacroExpander macros;
  private Map<Node, String> simpleTermTransformMap;
  private List<Node> listSimpleTermTransform;
  private Map<Node, String> simpleTermOrsimpleListTermTypes;

  private int popCount;

  GenerateAlternativeCodeForParser(File pkgDir, String aParsedAltName,
                                   String raParsedAltName,
                                   BufferedWriter file,
                                   ResolveTransformIds transformIds,
                                   ComputeCGNomenclature CG,
                                   ComputeSimpleTermPosition CTP,
                                   Map<Node, String> simpleTermTransformMap,
                                   MacroExpander macros,
                                   List<Node> listSimpleTermTransform,
                                   Map<Node, String> simpleTermOrsimpleListTermTypes)
  {
    this.pkgDir = pkgDir;
    this.file = file;
    currentAlt = aParsedAltName;
    realcurrentAlt = raParsedAltName;
    this.transformIds = transformIds;
    this.CG = CG;
    this.CTP = CTP;
    this.simpleTermTransformMap = simpleTermTransformMap;
    this.macros = macros;
    this.listSimpleTermTransform = listSimpleTermTransform;
    this.simpleTermOrsimpleListTermTypes = simpleTermOrsimpleListTermTypes;

    this.popCount = 0;
  }

  @Override
  public void caseAElem(final AElem node) {
    popCount++;
  }
  
  private String lookupInternalTypeName(Node term) {
    String name = simpleTermTransformMap.get(term);
    if(name != null)
    {
      return name;
    }
    else
    {
      return CG.getAltTransformElemTypes().get(term);
    }
  }
  
  private void generateVariableDeclaration(Node term) throws IOException {
    String typeName = lookupInternalTypeName(term);
    int position = CG.getTermNumbers().get(term);
    
    if(typeName.startsWith("L"))
    {
      macros.apply(file, "ParserListVariableDeclaration", new String[] {"" + position});
    }
    else if(typeName.equals("null"))
    {
      // No intermediate variable needed for null arguments
    }
    else
    {
      macros.apply(file, "ParserSimpleVariableDeclaration", new String[] {typeName, typeName.toLowerCase(), "" + position});
    }
  }

  @Override
  public void inAAltTransform(AAltTransform node)
  {
    for(PTerm term : node.getTerms())
    {
      try
      {
        generateVariableDeclaration(term);
      }
      catch(IOException e)
      {
        throw new RuntimeException("An error occured while writing to " +
                                   new File(pkgDir, "Parser.java").getAbsolutePath(), e);
      }
    }
  }

  @Override
  public void outAAltTransform(AAltTransform node)
  {
    int position;

    try
    {
      for(PTerm term : node.getTerms())
      {
        String type_name = lookupInternalTypeName(term);

        position = CG.getTermNumbers().get(term);

        if(type_name.startsWith("L"))
        {
          type_name = "list";
        }
        else if(type_name.equals("null"))
        {
          type_name = "null";
        }
        else
        {
          type_name = type_name.toLowerCase();
        }
        macros.apply(file, "ParserNewBodyListAdd", new String[] {type_name, "" + position});

      }
      if (popCount > 0) {
        // The nodeArrayList variables are numbered starting at 1, so the first popped variable has the number popCount and not popCount-1.
        macros.apply(file, "ParserNewCheck", new String[] {String.valueOf(popCount)});
      }
      macros.apply(file, "ParserNewTail");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  private void generateNewTermDeclarations(List<PTerm> params)
  {
    try
    {
      macros.apply(file, "ParserBraceOpening");
      for(PTerm term : params)
      {
        generateVariableDeclaration(term);
      }
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  private void generateSimpleTerm(Node node, TId id, TId simpleTermTail)
  {
    try
    {
      String type_name = lookupInternalTypeName(node);
      int position = CG.getTermNumbers().get(node).intValue();
      String termKey = currentAlt + "." + id.getText();
      int elemPosition = CTP.elems_position.get(termKey);
      int positionMap = 0;

      if(simpleTermTail != null)
      {
        // FIXME This condition is always false, because listSimpleTermTransform contains ASimpleTerm/ASimpleListTerm and not String! Is it safe to remove this?
        if(!listSimpleTermTransform.contains(id.getText()))
        {
          String type = CTP.positionsMap.get(realcurrentAlt + "." + id.getText());
          List<String> list = transformIds.getProdTransformIds().prod_transforms.get(type);
          if(list.indexOf(simpleTermTail.getText()) >= 0)
          {
            positionMap = list.indexOf(simpleTermTail.getText());
          }
        }

        if(simpleTermOrsimpleListTermTypes.get(node) != null)
        {
          String type = simpleTermOrsimpleListTermTypes.get(node);
          List<String> list = transformIds.getProdTransformIds().prod_transforms.get(type);
          if(list.indexOf(simpleTermTail.getText()) >= 0)
          {
            positionMap = list.indexOf(simpleTermTail.getText());
          }
        }
      }

      String type;
      if(type_name.startsWith("L"))
      {
        type_name = "list";
        type = "LinkedList";
      }
      else if(type_name.equals("null"))
      {
        type_name = "null";
        type = "Object";
      }
      else
      {
        type = type_name;
      }

      macros.apply(file, "ParserSimpleTerm", new String[]
                   {
                     type_name.toLowerCase(), ""+position,
                     type, ""+elemPosition, ""+positionMap
                   }
                  );
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void inASimpleTerm(ASimpleTerm node)
  {
    generateSimpleTerm(node, node.getId(), node.getSimpleTermTail());
  }

  @Override
  public void inASimpleListTerm(ASimpleListTerm node)
  {
    generateSimpleTerm(node, node.getId(), node.getSimpleTermTail());
  }

  @Override
  public void inANewTerm(ANewTerm node)
  {
    generateNewTermDeclarations(node.getParams());
  }

  @Override
  public void inANewListTerm(ANewListTerm node)
  {
    generateNewTermDeclarations(node.getParams());
  }

  @Override
  public void inAListTerm(AListTerm node)
  {
    try
    {
      macros.apply(file, "ParserBraceOpening");
      for(Node term : node.getListTerms())
      {
        generateVariableDeclaration(term);
      }
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void outAListTerm(AListTerm node)
  {
    try
    {
      int listPosition = CG.getTermNumbers().get(node);

      for(PListTerm listTerm : node.getListTerms())
      {
        String type_name = lookupInternalTypeName(listTerm);
        int position = CG.getTermNumbers().get(listTerm);

        if(!type_name.equals("null"))
        {
          if(type_name.startsWith("L"))
          {
            macros.apply(file, "ParserTypedLinkedListAddAll", new String[] {"list", ""+listPosition, "list", ""+ position});
          }
          else
          {
            macros.apply(file, "ParserTypedLinkedListAdd", new String[] {"list", ""+listPosition, type_name.toLowerCase(), ""+ position});
          }
        }
      }
      macros.apply(file, "ParserBraceClosing");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  private void generateNewTerm(Node node, AProdName prodName, List<PTerm> params)
  {
    String type_name = lookupInternalTypeName(node);
    int position = CG.getTermNumbers().get(node);
    String newAltName = name(prodName);

    try
    {
      if(type_name.startsWith("L"))
      {
        type_name = "list";
      }
      else
      {
        type_name = type_name.toLowerCase();
      }
      macros.apply(file, "ParserNewBodyNew", new String[] {type_name, ""+position, newAltName});

      if(!params.isEmpty())
      {
        String isNotTheFirstParam = "";

        for(PTerm term : params)
        {
          type_name = lookupInternalTypeName(term);
          position = CG.getTermNumbers().get(term);

          if(type_name.equals("null"))
          {
            macros.apply(file, "ParserNew&ListBodyParamsNull", new String[] {isNotTheFirstParam+"null"});
          }
          else
          {
            if(type_name.startsWith("L"))
            {
              type_name = "list";
            }
            else
            {
              type_name = type_name.toLowerCase();
            }
            macros.apply(file, "ParserNew&ListBodyParams", new String[] {isNotTheFirstParam+type_name, ""+position});
          }

          isNotTheFirstParam = ", ";
        }
      }
      macros.apply(file, "ParserNewBodyNewTail");
      macros.apply(file, "ParserBraceClosing");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "TokenIndex.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void outANewTerm(ANewTerm node)
  {
    generateNewTerm(node, (AProdName)node.getProdName(), node.getParams());
  }

  @Override
  public void outANewListTerm(ANewListTerm node)
  {
    generateNewTerm(node, (AProdName)node.getProdName(), node.getParams());
  }

  public String name(AProdName node)
  {
    if(node.getProdNameTail() != null)
    {
      return "A" +
             ResolveIds.name(node.getProdNameTail().getText()) +
             ResolveIds.name(node.getId().getText());
    }
    return "A" + ResolveIds.name(node.getId().getText());
  }

}
